# -*- coding: latin-1 -*-

# Python
from DateTime import DateTime
import email, re
import email.Header
from email import Utils

first_line_match = re.compile(r'(\S.*)')

#########
# Helpers

def tryDecoding(st, codings):
    for coding in codings:
        try:
            return st.decode(coding)
        except UnicodeError:
            pass
    raise UnicodeError, 'No suitable encoding found'


def wordWrap(text, width=72, prefix='>'):
    """
    Stole most of this code from:
    http://mail.zope.org/pipermail/zope/2002-August/120022.html
    Saved me a 0.5 hour. Thank you guys!
    """
    remaining = text.replace('\r\n', '\n')
    wrapped = []
    a = lambda st: wrapped.append(prefix + st)
    width = width - len(prefix)
    while len(remaining) > width:
        cut = width
        newline = remaining.find('\n', 0, cut)
        if newline != -1:
            cut = newline
        elif remaining[cut] != ' ':            
            temp = remaining.rfind(' ', 0, cut-1)
            if temp == -1:temp = remaining.find(' ', cut-1, len(remaining))
            if temp == -1: temp = len(remaining)
            cut = temp
        remaining_cut = remaining[:cut]
        if remaining_cut:
            wrapped.append(prefix + remaining_cut)
        else:
            wrapped.append('')
        remaining = remaining[cut+1:]
    if remaining:
        wrapped.append(prefix + ('\n'+prefix).join(remaining.split('\n')))
    return '\n'.join(wrapped)


class Attachment:

    "Description of attachment"

    __allow_access_to_unprotected_subobjects__=1

    def __init__(self, number, filename, content_type):
        self.number = number
        self.filename = filename
        self.content_type = content_type


_marker = []

class MessageWrapper:

    """
    A wrapper around a python email message from the python library, which 
    returns a message that is more suitable for web presentation.
    Especially strings are converted to site encoding.
    """

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, obj, encoding='utf-8'):
        self._encoding = encoding
        self.__obj = obj

    def __getattr__(self, attr, default=_marker):
        if default is _marker:
            return getattr(self.__obj, attr)
        else:
            return getattr(self.__obj, attr, default)

    def __getitem__(self, key):
        return self.get(key)

    def get(self, key, default=None):
        if not default is None:
            return self.__obj.get(key, default)
        else:
            return self.__obj.get(key)

    def _headerToDefaultCharset(self, header):
        "Returns an encoded header"
        encoding = self._encoding
        try:
            parts = email.Header.decode_header(header)
            new_header = email.Header.make_header(parts)
            return unicode(new_header).encode(encoding)
        except:
            try:
                return tryDecoding(header, ['utf-8', 'latin-1']).encode(encoding)
            except:
                return 'ENCODING ERRORS!'

    def Date(self):
        "Returns the message date as a DateTime"
        date = self.get('Date', None)
        if date is None:
            date = DateTime()
        else:
            try:
                date = DateTime(*Utils.parsedate(date)[:6])
            except:
                date = DateTime(1970,1,1)
        return date

    ## these generates html anchor tags from adresses

    def _mailAnchor(self, adress, name=''):
        "Returns email and [name] as a html anchor tag"
        if not adress and not name:
            return '' # The international sign of missing information :-)
        if not name:
            name = adress
        return '<a href="mailto:%s">%s</a>' % (adress, name)

    def _mailAnchors(self, adresses):
        "Converts [(name, adresses)] to html anchors"
        result = []
        for name, adress in adresses:
            result.append(self._mailAnchor(adress, name))
        return ', '.join(result)

    def anchorsFrom(self):
        "Returns sender name and email anchor"
        return self._mailAnchors(self.getFrom())

    def anchorsTo(self):
        "Returns recipients name and email anchor"
        return self._mailAnchors(self.getTo())

    def anchorsCc(self):
        "Returns recipients name and email anchor"
        return self._mailAnchors(self.getCc())

    ## these generates [(name, address)]

    def _getAdressList(self, key):
        "Returns an adress list"
        val = [self.get(key, '')]
        adresses = Utils.getaddresses(val)
        vals = []
        for name, adress in adresses:
            name = self._headerToDefaultCharset(name)
            vals.append((name, adress))
        return vals

    def getFrom(self):
        "Returns From persons"
        return self._getAdressList('From')

    def getTo(self):
        "Returns To persons"
        return self._getAdressList('To')

    def getCc(self):
        "Returns Cc persons"
        return self._getAdressList('Cc')

    ## returns mails and names in site encoding

    def _getAdressStr(self, key):
        "Returns adresses as 'email <name>, email <name>'"
        adresses = self._getAdressList(key)
        result = []
        for adress in adresses:
            result.append(Utils.formataddr(adress))
        return ', '.join(result)

    def getToStr(self):
        "Returns To persons as comma seperated string"
        return self._getAdressStr('To')

    def getFromStr(self):
        "Returns From persons as comma seperated string"
        return self._getAdressStr('From')

    def getCcStr(self):
        "Returns Cc persons as comma seperated string"
        return self._getAdressStr('Cc')

    def getComposeMails(self, key, compose_mode):
        """
        Returns adresses depending on the mode of the compose interface. Eg if 
        compose_mode=='reply' the orignial sender is returned as 'To' instead of 
        'From' etc.
        """
        # if we compose a new message, nothing is filled out
        if compose_mode=='compose':
            return ''
        elif compose_mode=='reply':
            return 'reply'
        elif compose_mode=='reply_all':
            return 'reply_all'

    ##

    def getText(self):
        "Returns the text of the message in the site encoding"
        text_parts = []
        mail = self.__obj
        for part in mail.walk():
            content_type = part.get_content_type()
            if content_type == 'text/plain':
                text_part = part.get_payload(decode=1)
                # apparently some mailers set charset to ascii, but encode as latin-1
                # so we cannot trust the charset. (Outlook express)
                try: 
                    cs = part.get_content_charset() # XXX iffy?
                    if cs:
#                        text_part = text_part.decode(cs).encode(self._encoding)
                        try:
                            u_text = unicode(text_part, cs, 'replace')
                        except LookupError:
                            u_text = repr(text_part) #u'Unknown charset %s' % cs
                        text_part = u_text.encode(self._encoding)
                except UnicodeDecodeError:
                    text_part = text_part.decode('latin-1').encode(self._encoding)
                if text_part:
                    text_parts.append(text_part)
        return '<hr>'.join(text_parts)


    def getWrappedText(self):
        return wordWrap(self.getText())


    def Subject(self):
        "Returns subject in site encoding"
        subject = self.__obj.get('Subject', '[No Subject]')
        return self._headerToDefaultCharset(subject)


    def hasAttachments(self) :
        "returns 1 if the email has attachements"
        mail = self.__obj
        return mail.is_multipart() or (
               mail.get_type() == 'application/octet-stream')


    def getAttachments(self):
        "Returns a list describing the attachments"
        result = []
        mail = self.__obj
        not_attachments = [None, 'text/plain', 'multipart/mixed', 'multipart/alternative', 'message/rfc822']
        if mail.is_multipart():
            number = 0
            for part in mail.walk():
                number += 1
                content_type = part.get_content_type()
                if not content_type in not_attachments:
                    filename = part.get_param('name', "attachment %s" % number)
                    filename = self._headerToDefaultCharset(filename)
                    result.append(Attachment(number, filename, content_type))
        else:
            content_type = mail.get_type()
            if content_type == 'application/octet-stream':
                filename = mail.get_filename()
                result.append(Attachment(1, filename, content_type))
        return result


    def getMessageId(self):
        "Returns the id of the message"
        return self.get('Message-Id', '')


    def wrapper_from_str(st):
        mail = email.message_from_string(st)
        return MessageWrapper(mail)
    wrapper_from_str = staticmethod(wrapper_from_str)

