# -*- coding: latin-1 -*-

# Python
from DateTime import DateTime
#import email

# Zope
from AccessControl import ClassSecurityInfo
from BTrees.IOBTree import IOBTree 

# CMF
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.CMFCorePermissions import View, ListFolderContents
from Products.CMFCore.CMFCorePermissions import setDefaultRoles

# Archetypes
from Products.Archetypes.public import BaseBTreeFolder, registerType

# Product
from schemas import message_folder_schema
from config import PROJECTNAME, VIEW_CONTENTS_PERMISSION,\
                   EDIT_CONTENTS_PERMISSION, ADD_CONTENTS_PERMISSION
from mxmImapClient import pathFromHex
from MessageWrapper import MessageWrapper

# message keys for the cached headers

MSG_ID = 'Uid'
MSG_SUBJECT = 'Subject'
MSG_DATE = 'Date'
MSG_IN_REPLY_TO = 'In-Reply-To'
MSG_FROM = 'From'
MSG_FROM_ANCHOR = 'FromAnchor'
MSG_TO = 'To'


# views for templates of same name, to remove any logic from zpt


class MxmPopUp_common:

    "Common methods"
    
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, context):
        self._context = context

    def portal_url(self):
        return self._context.portal_url.getPortalObject().absolute_url()


class MxmPopUp_waitView(MxmPopUp_common):

    def nexPage(self):
        "The next page"
        return self.REQUEST.get('next_page', '')

    def nexPageUrl(self):
        "The next page"
        context = self._context
        next_page = context.REQUEST.get('next_page', '')
        return '%s/%s' % (context.absolute_url(), next_page)


class MxmPopUp_updateFolderView(MxmPopUp_common):

    def __init__(self, context):
        self._context = context
        imapClient = context.getMxmMessageClient()
        idsToAdd, idsToRemove = imapClient.getUpdatedMessagesUids(context)
        # Put new messages uids in session
        idsToAdd.sort()
        context.putSessionUids(idsToAdd)
        context.putTotalMsgCount(len(idsToAdd))
        
        # remove uids that are no longer on the server
        self.idsToRemove = idsToRemove
        context.delCaches(idsToRemove)



class MxmPopUp_fetchBatchView(MxmPopUp_common):

    def view_url(self):
        return '%s/%s' % (self.absolute_url(), 'mxmPopUp_fetchBatch')

    def msgRangeLow(self):
        context = self._context
        return context.getTotalMsgCount() - context.newMessagesCount()

    def msgRangeHigh(self):
        context = self._context
        batchSize = context.getBatchSize()
        newMessagesCount = context.newMessagesCount()
        if batchSize >= newMessagesCount:
            batchSize = newMessagesCount
        return self.msgRangeLow() + batchSize

            


class mxmImapFolder(BaseBTreeFolder):

    content_icon = 'folder_icon.gif'

    global_allow = 0

    schema = message_folder_schema

    allowed_content_types = ('mxmImapFolder', 'mxmImapMessage')

    actions = (
        {
            'id': 'view',
            'name': 'View',
            'action': 'string:${object_url}/mxmImapFolder_view',
            'permissions': (View,)
        },
        {
            'id': 'metadata',
            'name': 'Properties',
            'action': 'string:${object_url}/base_metadata',
            'permissions': (EDIT_CONTENTS_PERMISSION,),
            'visible':0,
        },
    )


    def __init__(self, *args, **kwargs):
        BaseBTreeFolder.__init__(self, *args, **kwargs)
        self._caches = IOBTree()

    # Standard security settings
    security = ClassSecurityInfo()
    security.declareObjectProtected(View)

    ################
    # View factories.

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getMxmPopUp_waitView')
    def getMxmPopUp_waitView(self):
        "Returns a mxmPopUp_wait view"
        return MxmPopUp_waitView(self)

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getMxmPopUp_updateFolderView')
    def getMxmPopUp_updateFolderView(self):
        "Returns a UpdateFolderPopupView"
        return MxmPopUp_updateFolderView(self)

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getMxmPopUp_fetchBatchView')
    def getMxmPopUp_fetchBatchView(self):
        "Returns a UpdateFolderPopupView"
        return MxmPopUp_fetchBatchView(self)


    #######
    # Logic

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getMxmMessageFolder')
    def getMxmMessageFolder(self):
        "Returns self"
        return self

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getFolderName')
    def getFolderName(self):
        "Returns the foldername"
        if hasattr(self, 'folder_name'):
            return self.folder_name
        else: # this isn't really nessecary except for upgraded folders
            hexPath = self.getMxmMessageClient().getMessagePath(self)
            return '/'.join(pathFromHex(hexPath))

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getFolderDelimiter')
    def getFolderDelimiter(self):
        "Returns the folder delimiter"
        return getattr(self, 'folder_delimiter', '/')

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getDepth')
    def getDepth(self):
        "Returns the depth of the folder from the messageClient"
        return self.getMxmMessageClient().calculateDepth(self)

    
    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'isHidden')
    def isHidden(self):
        "Return true if folder is hidden"
        return self.folderPath() in self.getMxmMessageClient().getHiddenFolders()

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getMessageFolders')
    def getMessageFolders(self, depth=0, showHidden=1):
        "Recursive folder search"
        result = []
        folders = self.objectValues('mxmImapFolder')
        decorated = [(f.title_or_id().lower(), f) for f in folders]
        decorated.sort()
        sorted = [d[-1] for d in decorated]
        for folder in sorted:
            if showHidden or not folder.isHidden():
                result.append(folder)
                result += folder.getMessageFolders(depth=depth+1, showHidden=showHidden)
        return result

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'folderPath')
    def folderPath(self):
        "Returns a relative folder path"
        return self.getMxmMessageClient().getFolderPathJoined(self)
    
    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getAttachment')
    def getAttachment(self, uid, number):
        "Returns the attachment with number"
        folderPath = self.folderPath()
        return self.getMxmMessageClient().getAttachment(folderPath, uid, number)


    #########################
    # cache handler methods
    # A message/header/cache is a simple dict with these keys:
    # MSG_ID, MSG_SUBJECT, MSG_DATE, MSG_IN_REPLY_TO, MSG_FROM, MSG_TO

    # debugging
    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'clearCache')
    def clearCache(self):
        "Clears the cache"
        self._caches = IOBTree()
        return 'Cache cleared'

    def addCache(self, cache):
        "A message header/cache is a simple dict"
        self._caches[cache[MSG_ID]] = cache

    def addCaches(self, caches):
        "Add a list of message caches"
        for cache in caches:
            self.addCache(cache)

    def delCache(self, id):
        "Delete a message cache by id"
        del self._caches[id]

    def delCaches(self, ids):
        "Delete a list of message caches by id"
        for id in ids:
            self.delCache(id)

    def moveCaches(self, ids, target):
        caches = self.getCaches(ids)
        target.addCaches(caches)
        self.delCaches(ids)

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getCache')
    def getCache(self, id):
        return self._caches[id]

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getCaches')
    def getCaches(self, ids):
        "Returns list of caches from ids"
        _caches = self._caches
        return [_caches[id] for id in ids if _caches.has_key(id)]

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getAllCaches')
    def getAllCaches(self):
        "Returns list of caches"
        return self._caches.values()

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getCacheUids')
    def getCacheUids(self):
        "Returns list of caches"
        return self._caches.keys()

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getCacheCount')
    def getCacheCount(self):
        "Returns number of messages"
        return len(self._caches)

    def setCacheFromMsg(self, id, msg):
        "Sets the cache values from an email message"
        wrap = MessageWrapper.wrapper_from_str(msg)
        cache = {
            MSG_ID : id,
            MSG_SUBJECT : wrap.Subject(),
            MSG_DATE : wrap.Date(),
            MSG_IN_REPLY_TO : wrap.get('In-Reply-To', ''),
            MSG_FROM : wrap.getFrom(),
            MSG_FROM_ANCHOR : wrap.anchorsFrom(),
            MSG_TO : wrap.getTo(),
        }
        self.addCache(cache)

    def setCachesFromMessages(self, messageItems):
        "Sets the cache values from a list of email messages"
        msgDict = {}
        for id, msg in messageItems:
            wrap = MessageWrapper.wrapper_from_str(msg)
            cache = {
                MSG_ID : id,
                MSG_SUBJECT : wrap.Subject(),
                MSG_DATE : wrap.Date(),
                MSG_IN_REPLY_TO : wrap.get('In-Reply-To', ''),
                MSG_FROM : wrap.getFrom(),
                MSG_FROM_ANCHOR : wrap.anchorsFrom(),
                MSG_TO : wrap.getTo(),
            }
            msgDict[id] = cache
        self.updateCaches(msgDict)

    def updateCaches(self, messages):
        "Update the cache from a plain mesage dict"
        return self._caches.update(messages)

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getUnreadMessageCount')
    def getUnreadMessageCount(self):
        "Returns list of messages"
        return len(self._caches)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getHighestId')
    def getMaxId(self):
        """
        Messages are stored with integers as keys. These are the same as the UID 
        from the server. It can be used for synchronisation.
        """
        oids = self._caches.keys()
        if not oids:
            return 0
        return max(oids)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getAllCaches')
    def getSortedCaches(self, key, reverse=0):
        "Returns list of caches, sorted by key"
        caches = self.getAllCaches()
        decorated = [(cache[key], cache) for cache in caches]
        decorated.sort()
        sorted = [cache[-1] for cache in decorated]
        if reverse:
            sorted.reverse()
        return sorted


    ###################################
    # actions for the syncronize popups

    IDS_TO_ADD_KEY = 'mxm_ids_to_add'

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'session')
    def session(self):
        return self.REQUEST.SESSION
    
    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'putSessionUids')
    def putSessionUids(self, uids):
        "Saves a list if integer uids in the session object"
        self.session()[self.IDS_TO_ADD_KEY] = uids

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getSessionUids')
    def getSessionUids(self):
        "Gets a list if integer uids from the session object"
        return self.session().get(self.IDS_TO_ADD_KEY, [])

    DEFAULT_BATCH_SIZE = 30
    
    def getBatchSize(self):
        return self.REQUEST.get('batch_size', self.DEFAULT_BATCH_SIZE)


    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getUidBatch')
    def getUidBatch(self):
        "Returns next batch of uids"
        uids = self.getSessionUids()
        batch_size = self.getBatchSize()
        batch = uids[:batch_size]
        self.putSessionUids(uids[batch_size:])
        return batch

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'newMessagesCount')
    def newMessagesCount(self):
        return len(self.session().get(self.IDS_TO_ADD_KEY, []))

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'fetchBatchAction')
    def fetchBatchAction(self, REQUEST=None):
        "Does actual fetching of batch"
        uids = self.getUidBatch()
        redirect = REQUEST.RESPONSE.redirect
        if uids:
            imapClient = self.getMxmMessageClient()
            self.syncHeadersFromUids(self, uids)
            redirect(self.absolute_url() + '/mxmPopUp_fetchBatch')
        else:
            redirect(self.absolute_url() + '/mxmPopUp_fetchDone')

    MESSAGE_COUNT_KEY = 'mxm_total_message_count'
    
    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'putTotalMsgCount')
    def putTotalMsgCount(self, msgCount):
        "Puts the total message count into session var"
        self.session()[self.MESSAGE_COUNT_KEY] = msgCount

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getTotalMsgCount')
    def getTotalMsgCount(self):
        "Gets the total message count from session var"
        return self.session().get(self.MESSAGE_COUNT_KEY, 0)

    # message methods

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getMessage')
    def getMessage(self, uid):
        "Returns a message from the server"
        folderName = self.getFolderName()
        msg = self.getMxmMessageClient().fetchMessage(folderName, uid)
        return msg

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'getMessageWrapper')
    def getMessageWrapper(self, uid):
        "Returns a wrapped message from the server"
        return MessageWrapper(self.getMessage(uid))

    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'markAsRead')
    def markAsRead(self, uid):
        "Marks the message as read"
        uid = int(uid)
        self.getMxmMessageClient().markAsRead(self, uid)

    
    security.declareProtected(VIEW_CONTENTS_PERMISSION, 'isUnread')
    def isUnread(self, uid):
        "Returns true if message with id is unread"
        return self.getMxmMessageClient().isUnread(self, uid)


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'deleteMessages')
    def deleteMessages(self, ids=None, REQUEST=None):
        """
        Delete messages by id
        """
        if ids != None:
            # delete the caches from this folder
            len_Ids = len(ids)
            ids = map(int, ids)
            self.delCaches(ids)
            # delete them from the mail server
            folderName = self.getFolderName()
            self.getMxmMessageClient().deleteMessages(folderName, ids)
            message = '%s messages deleted' % len_Ids
        else:
            message = 'No messages deleted'
        if REQUEST:
            REQUEST.RESPONSE.redirect(self.absolute_url() + '?portal_status_message=%s' % message)


    security.declareProtected(EDIT_CONTENTS_PERMISSION, 'moveMessages')
    def moveMessages(self, to_folder_path, ids=None, REQUEST=None):
        """
        Move messages to another folder
        to_folder_path is the path to the local folder
        """
        if ids != None:
            # Move the cached/local messages
            len_Ids = len(ids)
            ids = map(int, ids)
            to_folder = self.unrestrictedTraverse(to_folder_path)
            self.moveCaches(ids, to_folder)
            # delete them from the mail server
            to_folder_name = to_folder.getFolderName()
            from_folder_name = self.getFolderName()
            self.getMxmMessageClient().moveMessages(from_folder_name, ids, to_folder_name)
            message = '%s messages moved' % len_Ids
        else:
            message = 'No messages moved'
        if REQUEST:
            REQUEST.RESPONSE.redirect(self.absolute_url() + '?portal_status_message=%s' % message)


setDefaultRoles( ADD_CONTENTS_PERMISSION, ('Manager', 'Owner',))
setDefaultRoles(EDIT_CONTENTS_PERMISSION, ('Manager', 'Owner',))
setDefaultRoles(VIEW_CONTENTS_PERMISSION, ('Manager', 'Owner',))

registerType(mxmImapFolder, PROJECTNAME)

